@@ -1,5 +1,34 @@
 #include "config.h"
 
+#ifdef __linux__
+/* Three kludges for the price of one; glibc's fts package doesn't
+   support LFS, so fall back to nftw. Do it here, before the headers
+   define _BSD_SOURCE and set the prefer BSD behavior, preventing us
+   from including XPG behavior with nftw later on. Check for linux and
+   not glibc for the same reason. Can't include features.h without
+   prefering bsd.
+
+   Once glibc has a working fts, remove nftw completely along with
+   this hack.
+ */
+
+#undef HAVE_FTS_OPEN
+#define _GNU_SOURCE
+#endif
+
+#if 0
+/* the fts function does not like 64bit-on-32bit, but we don't need it here anyway. */
+#ifdef _FILE_OFFSET_BITS
+#undef _FILE_OFFSET_BITS
+#define LARGE_FILES_ARE_ENABLED 1
+#endif
+#ifdef _LARGE_FILES
+#undef _LARGE_FILES
+#define LARGE_FILES_ARE_ENABLED 1
+#endif
+#endif // 0
+
+#include <assert.h>
 #include <errno.h>
 #include <fcntl.h>
 #include <stdlib.h>
@@ -15,32 +44,21 @@
 
 #include "srm.h"
 
-#ifndef FTS_D
-#define FTS_D      0
-#define FTS_DC     1
-#define FTS_DNR    2
-#define FTS_DOT    3
-#define FTS_DP     4
-#define FTS_ERR    5
-#define FTS_F      6
-#define FTS_SL     7
-#define FTS_SLNONE 8
-#define FTS_NS     9 
-#endif /* ndef FTS_D */
-
 int prompt_user(const char *string) {
   char inbuf[8];
+  assert(string);
 
-  printf("%s", string);
+  printf("%s", string); fflush(stdout);
   fgets(inbuf, 4, stdin);
   return strncmp(inbuf, "y", 1) == 0;
 }
 
 int check_perms(const char *path) {
   int fd;
+  assert(path);
 
-  if ( ((fd = open(path, O_WRONLY)) == -1) && (errno == EACCES) ) {
-    if ( chmod(path, S_IRUSR | S_IWUSR) == -1 ) {
+  if ( ((fd = open(path, O_WRONLY)) < 0) && (errno == EACCES) ) {
+    if ( chmod(path, S_IRUSR | S_IWUSR) < 0 ) {
       errorp("Unable to reset %s to writable (probably not owner) "
 	     "... skipping", path);
       return 0;
@@ -52,14 +70,12 @@ int check_perms(const char *path) {
 }
 
 int prompt_file(const char *path, int fts_flag) {
-  int fd, return_value = 1;
-  size_t bufsize = strlen(path) + 80;
+  int fd=-1, return_value = 1;
+  size_t bufsize;
   char *buf;
+  assert(path);
 
-  if ( (buf = (char *)malloc(bufsize)) == NULL ) {
-    errorp("Out of memory at line %d in prompt_file()", __LINE__);
-    return 0;
-  }
+  bufsize = strlen(path) + 80;
 
   if (options & OPT_F) {
     if (options & OPT_V)
@@ -67,7 +83,12 @@ int prompt_file(const char *path, int fts_flag) {
     return check_perms(path);
   }
 
-  if ( (fts_flag != FTS_SL) && ((fd = open(path, O_WRONLY)) == -1) &&
+  if ( (buf = (char *)alloca(bufsize)) == NULL ) {
+    errorp("Out of memory at line %d in prompt_file()", __LINE__);
+    return 0;
+  }
+
+  if ( (fts_flag != FTS_SL) && ((fd = open(path, O_WRONLY)) < 0) &&
        (errno == EACCES) )
     {
       /* Not a symlink, not writable */
@@ -85,40 +106,83 @@ int prompt_file(const char *path, int fts_flag) {
   if ((options & OPT_V) && return_value) 
     printf("removing %s\n", path);
 
-  free(buf);
-  close(fd); /* close if open succeeded, or silently fail */
+  if(fd >= 0)
+    close(fd); /* close if open succeeded, or silently fail */
 
   return return_value;
 }
 
 int process_file(char *path, int flag) {
-  while (path[strlen(path) - 1] == '/') 
+  assert(path);
+
+  while (path[strlen(path) - 1] == DIRSEP) 
     path[strlen(path)- 1] = '\0';
 
   switch (flag) {
+#ifdef FTS_D
   case FTS_D: break;
+#endif
+
+#ifdef FTS_DC
   case FTS_DC:
     error("cyclic directory entry %s", path);
     break;
+#endif
+
+#ifdef FTS_DNR
   case FTS_DNR:
+#endif
     error("%s: permission denied", path);
     break;
+
+#ifdef FTS_DOT
   case FTS_DOT: break;
+#endif
+
+#ifdef FTS_DP
   case FTS_DP:
     if (options & OPT_R) {
-      if ( prompt_file(path, flag) && (rename_unlink(path) == -1) )
+      if ( prompt_file(path, flag) && (rename_unlink(path) < 0) )
 	errorp("unable to remove %s", path);
     } else {
       error("%s is a directory", path);
     }
     break;
+#endif
+
+#ifdef FTS_ERR
   case FTS_ERR:
     error("fts error on %s", path);
     break;
+#endif
+
+#ifdef FTS_NS
+  case FTS_NS:
+#endif
+#ifdef FTS_NSOK
+  case FTS_NSOK:
+#endif
+    /* if we have 32bit system and file is >2GiB the fts functions can not stat them, so we just ignore the fts error. */
+#ifndef LARGE_FILES_ARE_ENABLED
+    /* Ignore nonexistant files with -f */
+    if ( !(options & OPT_F) )
+      {
+	error("unable to stat %s", path);
+	break;
+      }
+#endif
+    /* no break here */
+
+#ifdef FTS_F
   case FTS_F:
+#endif
+#ifdef FTS_SL
   case FTS_SL:
+#endif
+#ifdef FTS_SLNONE
   case FTS_SLNONE:
-    if ( prompt_file(path, flag) && (sunlink(path, options & OPT_S) == -1) ) {
+#endif
+    if ( prompt_file(path, flag) && (sunlink(path, options & SUNLINK_MODE_MASK) < 0) ) {
       if (errno == EMLINK) 
 	error("%s has multiple links, this one has been removed but not "
 	      "overwritten", path);
@@ -126,11 +190,7 @@ int process_file(char *path, int flag) {
 	errorp("unable to remove %s", path);
     }
     break;
-  case FTS_NS:
-    if ( !(options & OPT_F) ) /* Ignore nonexistant files with -f */
-      error("unable to stat %s", path);
-  default:
-    break;
+
   }
   return 0;
 }
@@ -141,19 +201,24 @@ int tree_walker(char **trees) {
   FTSENT *current_file;
   FTS *stream;
   int i = 0;
-  
+
+  assert(trees);
+
   while (trees[i] != NULL) {
-    while (trees[i][strlen(trees[i]) - 1] == '/')
+    while (trees[i][strlen(trees[i]) - 1] == DIRSEP)
       trees[i][strlen(trees[i]) -1] = '\0';
     i++;
   }
 
   if ( (stream = fts_open(trees, FTS_PHYSICAL | FTS_NOCHDIR, NULL)) == NULL )
     errorp("fts_open() returned NULL");
-  while ( (current_file = fts_read(stream)) != NULL) {
-    process_file(current_file->fts_path, current_file->fts_info);
-    if ( !(options & OPT_R) )
-      fts_set(stream, current_file, FTS_SKIP);
+  else {
+    while ( (current_file = fts_read(stream)) != NULL) {
+      process_file(current_file->fts_path, current_file->fts_info);
+      if ( !(options & OPT_R) )
+	fts_set(stream, current_file, FTS_SKIP);
+    }
+    fts_close(stream);
   }
   return 0;
 }
@@ -164,13 +229,20 @@ int tree_walker(char **trees) {
 /* Shut up tru64's cc(1) */
 #define _XOPEN_SOURCE_EXTENDED
 #endif
+
+#define _GNU_SOURCE
 #include <ftw.h>
 
 int ftw_process_path(const char *opath, const struct stat *statbuf, int flag,
 		 struct FTW *dummy)
 {
-  size_t path_size = strlen(opath) + 1;
-  char *path = (char *)alloca(path_size);
+  size_t path_size;
+  char *path;
+
+  assert(opath);
+
+  path_size = strlen(opath) + 1;
+  path = (char *)alloca(path_size);
 
   if (path == NULL) {
     errno = ENOMEM;
@@ -206,9 +278,10 @@ int ftw_process_path(const char *opath, const struct stat *statbuf, int flag,
 
 int tree_walker(char **trees) {
   int i = 0;
+  assert(trees);
 
   while (trees[i] != NULL) { 
-    while (trees[i][strlen(trees[i]) - 1] == '/')
+    while (trees[i][strlen(trees[i]) - 1] == DIRSEP)
       trees[i][strlen(trees[i]) -1] = '\0';
     if (options & OPT_R)
       nftw(trees[i], ftw_process_path, 10, FTW_DEPTH);
@@ -219,6 +292,9 @@ int tree_walker(char **trees) {
   return 0;
 }
 
+#elif defined(_WIN32)
+/* code is in win/tree_walker.cpp */
+
 #else
 #error No tree traversal function found
 #endif
